#include <stdio.h>
#include <unistd.h>
#include "err.h"
#include <math.h>
#include <errno.h>
#include <string.h>

#include "tracker.h"

struct instrument instrument[MAXINSTR];
struct track track[MAXTRACK];
struct song song;

char *filename;
char *suffix = "";

u8 trackmap[MAXTRACK];
u16 trackoffs[MAXTRACK];
int ntrack = 0;
int ninstr = 0;
u8 lasttrack[6];

u8 instrmap[MAXINSTR];

u8 bits, bitcount = 0;
u8 xpos;
int bytecount = 0;

void newline() {
	if(xpos) printf("\n");
	xpos = 0;
}

void putbyte(u8 v) {
	if(xpos == 0) {
		printf("\t\tbyte\t");
	} else {
		printf(", ");
	}
	printf("$%02x", v);
	xpos++;
	if(xpos == 16) newline();
	bytecount++;
}

void putbit(int v) {
	bits >>= 1;
	if(v) bits |= 0x80;
	bitcount++;
	if(bitcount == 8) {
		putbyte(bits);
		bitcount = 0;
	}
}

void chunk(unsigned int v, int width) {
	int i;

	if(v >= (1 << width)) err(1, "%d doesn't fit in %d bits", v, width);
	for(i = width - 1; i >= 0; i--) {
		putbit(v & (1 << i));
	}
}

void pad() {
	while(bitcount) putbit(0);
}

void pad16() {
	pad();
	if(bytecount & 1) chunk(0, 8);
}

int main(int argc, char **argv) {
	int i, j, ch;

	if(argc > 2 && !strcmp(argv[1], "-s")) {
		suffix = argv[2];
		argc -= 2;
		argv += 2;
	}

	if(argc != 2) {
		err(1, "Usage: %s [-s suffix] songfile >song.s", argv[0]);
	}

	filename = argv[1];

	song.length = 1;
	if(loadfile()) {
		err(1, "Error loading %s", filename);
	}

	// Remap the tracks.

	for(i = 0; i < song.length; i++) {
		for(ch = 0; ch < 6; ch++) {
			trackmap[song.line[i].track[ch]] = 1;
		}
	}
	for(i = 1; i < MAXTRACK; i++) {
		if(trackmap[i]) trackmap[i] = ++ntrack;
	}
	trackmap[0] = 0;

	printf("\t\talign 2\n");

	printf("bf_song%s\n", suffix);

	for(i = 0; i < song.length; i++) {
		for(ch = 0; ch < 6; ch++) {
			if(song.line[i].track[ch] != lasttrack[ch]) {
				chunk(1, 1);
				chunk(trackmap[song.line[i].track[ch]], TRACKNUMBITS);
				lasttrack[ch] = song.line[i].track[ch];
			} else {
				chunk(0, 1);
			}
			if(song.line[i].transp[ch]) {
				chunk(1, 1);
				chunk(song.line[i].transp[ch] & ((1 << TRANSPBITS) - 1), TRANSPBITS);
			} else {
				chunk(0, 1);
			}
		}
	}
	for(ch = 0; ch < 6; ch++) {
		if(0 != lasttrack[ch]) {
			chunk(1, 1);
			chunk(0, TRACKNUMBITS);
		} else {
			chunk(0, 1);
		}
		if(ch == 5) {
			chunk(1, 1);
			chunk(1, 1);
		} else {
			chunk(0, 1);
		}
	}
	pad16();
	newline();

	printf("bf_tracks%s\n", suffix);

	printf("\t\tword\t0\t' Track 00 is a null pointer.\n");
	for(i = 0; i < ntrack; i++) {
		printf("\t\tword\t@:track_%02x\n", i + 1);
	}
	bytecount = 0;
	for(i = 0; i < MAXTRACK; i++) {
		if(trackmap[i]) {
			int lasti = -1, cutpos = 0;

			printf(":track_%02x\n", trackmap[i]);
			for(j = 0; j < TRACKLEN; j++) {
				if(track[i].line[j].instr || track[i].line[j].note) {
					cutpos = j + 1;
				}
			}
			if(cutpos > TRACKLEN - 3) cutpos = TRACKLEN;
			//if(cutpos != TRACKLEN) fprintf(stderr, "C %d\n", cutpos);
			for(j = 0; j < cutpos; j++) {
				if(track[i].line[j].instr) {
					int inum = track[i].line[j].instr;
					if(inum != lasti) {
						chunk(1, 1);
						chunk(inum, INSTRBITS);
						if(inum >= ninstr) ninstr = inum + 1;
						lasti = inum;
					} else {
						chunk(0, 1);
					}
				} else {
					chunk(0, 1);
				}
				if(track[i].line[j].note) {
					chunk(1, 1);
					chunk(track[i].line[j].note - 1, NOTEBITS);
				} else {
					chunk(0, 1);
				}
			}
			if(cutpos != TRACKLEN) {
				chunk(1, 1);
				chunk(0, INSTRBITS);
			}
			pad();
			newline();
		}
	}
	//pad16();
	//newline();

	for(i = 1; i < ninstr; i++) {
		for(j = 0; j < instrument[i].length; j++) {
			if((instrument[i].line[j] & 0xff00) == 0x0100) {
				int inum = (instrument[i].line[j] & 0x00ff);
				if(inum >= ninstr) ninstr = inum + 1;
			}
		}
	}

	printf("bf_instruments%s\n", suffix);

	printf("\t\tbyte\t0\t' Dummy slot for instrument 00.\n");
	for(i = 1; i < ninstr; i++) {
		printf("\t\tbyte\t(@:instr_%02x - @bf_instruments%s) / 2\n", i, suffix);
	}
	if(ninstr & 1) printf("\t\tbyte\t0\t' pad\n");
	for(i = 1; i < ninstr; i++) {
		bytecount = 0;
		printf(":instr_%02x\n", i);
		for(j = 0; j < instrument[i].length; j++) {
			chunk(instrument[i].line[j] >> 8, 4);
			chunk(instrument[i].line[j] & 255, 8);
		}
		if(!(instrument[i].length && ((instrument[i].line[instrument[i].length - 1] & 0xff00) == 0x0100))) {
			chunk(0, 4);
		}
		pad16();
		newline();
	}

	return 0;
}
