
			include "propeller.i"
			include "demo.i"

			org	0

			include	"unpack.i"

			' This is needed to get @freqbase right.
			horg	UNPACK_LIBREGS * 4

			' Initialize the physical EEPROM interface. At cog start, all pins are inputs.

ee_count		or	DIRA, scl			' SCL is an output, and will be held low.

			' Modify the bitstream reader to access EEPROM rather than hub RAM.
			' This works even if the current reader state is in the middle of a byte, because
			' hub RAM contents and EEPROM contents are identical at boot.

ee_last			movs	p_getbit, #eegetbyte		' initial ee_last must be != 0 and != @bf_song

ee_timer		jmp	#stagetwo

			' THigh		SCL high		600 ns		48
			' TLow		SCL low			1300 ns		104
			' TF		SCL/SDA rise/fall time	300 ns		24
			' THd:Sta	Start hold time		600 ns		48
			' TSu:Sto	Stop setup time		600 ns		48
			' TAA		Output valid from clock	900 ns		72
			' TBuf		Bus free time		1300 ns		104

			' 100 ns == 8 CNT cycles
eegetbyte
			mov	ee_timer, CNT
			add	ee_timer, #40			' Gonna need some time for the first instructions
			cmp	reader, ee_last		wz
		if_z	jmp	#eenext

			' Having read the final byte in a sequence, nak it,
			' then perform a repeated START.

			' In the special case of first access, we might be out
			' of sync with the EEPROM chip, so keep clocking out high
			' bits (nak) until the lines are in the proper state for
			' a start condition.

:nak			call	#eenak

			' SDL is low, SDA is high

			waitcnt	ee_timer, #72 + 8		' Wait for TLow - TF (80 + 8)
			or	OUTA, scl			' SCL high
			waitcnt	ee_timer, #24			' Wait for TAA (72 + 8)
			test	sda, INA		wc	' Verify that SDA is high
		if_nc	jmp	#:nak

			waitcnt	ee_timer, #48			' Wait for TBuf - TAA (24)
			or	DIRA, sda			' Pull SDA low while SCL is high => start
			waitcnt	ee_timer, #24			' Wait for THd:Sta (48)
			andn	OUTA, scl			' SCL low
			waitcnt	ee_timer, #80 + 8		' Wait for TF (24)

			mov	nextbyte, #$a0 << 1
			call	#eetransmitfirst
			mov	nextbyte, reader
			shr	nextbyte, #8
			call	#eetransmit
			mov	nextbyte, reader
			call	#eetransmit

			mov	ee_last, reader

			call	#eenak

			' SDL is low, SDA is high

			waitcnt	ee_timer, #104			' Wait for TLow - TF (80 + 8)
			or	OUTA, scl			' SCL high
			waitcnt	ee_timer, #48			' Wait for TBuf (104)
			or	DIRA, sda			' SDA low => repeated start
			waitcnt	ee_timer, #24			' Wait for THd:Sta (48)
			andn	OUTA, scl			' SCL low
			waitcnt	ee_timer, #80 + 8		' Wait for TF (24)

			mov	nextbyte, #$a1 << 1
			call	#eetransmitfirst

eenext			add	ee_last, #1

			mov	nextbyte, #$ff			' Bit 8 is ack for previous byte
			call	#eetransmit

			and	nextbyte, #$ff

			' After receiving a byte, SCL is held low, nextbyte
			' contains the byte, ee_last has been incremented, and
			' the EEPROM chip is waiting for the final ACK/NAK clock
			' pulse. If the next read is from the following EEPROM
			' address, we'll generate an ACK and keep reading.
			' Otherwise we'll generate a NAK, a repeated start,
			' transmit the new address, and read.

			jmp	#getbit_back

eenak
			mov	ee_count, #1
			andn	DIRA, sda			' SDA high = nak
			jmp	#eetransmit_2
eetransmitfirst
			mov	ee_count, #8
eetransmit
			test	nextbyte, #$100		wz
			muxz	DIRA, sda			' Write (low or floating)
eetransmit_2		waitcnt	ee_timer, #72 + 8		' Wait for TLow - TF (80 + 8)
			or	OUTA, scl			' SCL high
			waitcnt	ee_timer, #24 + 8		' Wait for TAA (72 + 8)
			test	sda, INA		wc	' Read
			rcl	nextbyte, #1
			andn	OUTA, scl			' SCL low
			waitcnt	ee_timer, #80 + 8		' Wait for TF (24 + 8)
			djnz	ee_count, #eetransmit
			mov	ee_count, #9
eenak_ret
eetransmitfirst_ret
eetransmit_ret		ret

scl			long	1 << 28
sda			long	1 << 29

			fit	UNPACK_LIBREGS + EEPROM_LIBREGS

stagetwo
			' The boot code unpacks this file into cog 0. The first
			' register contains a jump into the EEPROM setup routine
			' at the top of this file, which in turn jumps here.

			' Clear the palette and the playroutine variables

:clr			wrlong	zero, addr
			add	addr, #4
			cmp	addr, =@channels+CH_SIZEOF*6	wz
		if_ne	jmp	#:clr

			' Calculate the frequency table

:freq			rdword	t1, src
			add	src, #2
			shr	t1, octave
			wrword	t1, addr
			add	addr, #2
			cmp	src, =@freqbase + 24	wz
		if_z	sub	src, #24
		if_z	sub	octave, #1
			djnz	count, #:freq

			' The bit reader has been set up to read from EEPROM. Verify that
			' that we've been loaded from EEPROM, and not via the serial port.

			mov	savereader, reader
			mov	reader, #0
			mov	gc_count, #32
			call	#getchunk
			rev	gc_result, #0
			xor	0, gc_result
			mov	gc_count, #32
			call	#getchunk
			rev	gc_result, #0
			xor	1, gc_result

			' Registers 0 and 1 are zero if the first 8 bytes of
			' EEPROM (including checksum) match the start of cog
			' RAM. Load the vga driver etc. from hub RAM before
			' checking for this condition.
			
			mov	reader, savereader
			movs	p_getbit, #getbyte	' Read further bits from RAM rather than EEPROM. 

			' Unpack the vga program. It follows init in the bitfile.

			mov	up_dest, =@unpack_area - 1
			call	#unpack

			' Start the video generation
launch_vga
			wrword	black, =@palette_area	' This will also sync CNT with the hub cycle.
			mov	CNT, CNT
			add	CNT, =$16000
			wrlong	CNT, =@nextvbl
			coginit	l_vga6
			coginit	l_vga7

			waitcnt	CNT, =$16000
			waitcnt	CNT, =$48000
			waitcnt CNT, =$5080

			' Verify that the video cogs are in sync. Sometimes
			' their PLLs are slightly out of phase, and this error
			' somehow gets amplified into an 8 pixel jitter.
			' Instead of trying to ascertain and eliminate the
			' conditions that lead to this, I've chosen to just
			' measure the duration between horizontal sync pulses,
			' and if it's not exactly 2236 cycles, reboot the vga
			' cogs.

			mov	t1, #6
			lockclr t1 wc
			if_c	jmp #video_restart 'HAXXX
			mov	t1, #7
			lockclr t1 wc
			if_c	jmp #video_restart 'HAXXX
			jmp #video_ok


video_restart
			mov	t1, #6
			cogstop	t1
			mov	t1, #7
			cogstop	t1
			jmp	#launch_vga
video_ok
			or	0, 1		wz
		if_nz	jmp	#fromram

			' Launch the synthesizer. It follows vga in the bitfile.

			mov	up_dest, =@unpack_area - 1
			call	#unpack

			coginit	l_synth

			' Wait for the VGA monitor to lock on to our signal, and
			' to get some distance from the audible start-up click.

			' At the same time, measure the activity on the serial port
			' to see if we should run the hidden part.

			mov	CTRA, =%10_01110_000_00000000_000000_000_000000 + 31
			mov	FRQA, #1

			add	CNT, =$10000000
			waitcnt	CNT, #0

			cmp	threshold, PHSA		wc wz

		if_b	rdword	t1, =@filetableptr
		if_b	add	t1, #32 * 2
		if_b	wrword	t1, =@filetableptr

			' Launch the playroutine. It follows synth in the bitfile.

			mov	up_dest, =(UNPACK_LIBREGS + EEPROM_LIBREGS) * 4 - 1
			call	#unpack

			coginit	zero		' Replaces cog 0

fromram
			' Inform the user that the demo must be loaded from EEPROM.

			wrlong	palette, =@palette_area
			
			mov	DIRA, #$ff
			movi	CTRA, #%0_00001_111_00000000_000000_000_000000 >> 23
			movi	FRQA, #$10000000 >> 23
			mov	VSCL, =1
			movi	VCFG, #%0_01_1_00_000_00000000000_000_0_00000000 >> 23

			rdlong	t1, =@nextvbl
:w			rdlong	t2, =@nextvbl
			cmp	t1, t2			wz
		if_z	jmp	#:w

:vbl
			mov	VSCL, #1
			rdlong	t1, =@nextvbl
			waitcnt	t1, #0
			mov	VSCL, #20
			nop
			nop
			mov	VSCL, #71
			waitvid	zero, #0

			movs	VCFG, #$01

			mov	VSCL, =2540
			mov	y, #8 * 2
:lp			waitvid	zero, #0
			djnz	y, #:lp

			mov	txtaddr, =@message
			mov	count, #7
:row
			mov	addr, =$8000
			mov	y, #32
:vline
			mov	VSCL, =620
			waitvid	zero, #0

			mov	x, #15
			mov	VSCL, =16 << 12 | 256
:xlp
			rdbyte	t1, txtaddr
			add	txtaddr, #1
			shr	t1, #1		wc
			shl	t1, #7
			add	t1, addr
			rdlong	bits, t1
		if_c	shr	bits, #1
			waitvid	vidpal, bits
			djnz	x, #:xlp

			mov	VSCL, =620
			waitvid	zero, #0

			add	addr, #4
			sub	txtaddr, #15
			djnz	y, #:vline

			add	txtaddr, #15
			djnz	count, #:row

			movs	VCFG, #0
			jmp	#:vbl

eeprom_ok

addr			long	@palette_area
src			long	@freqbase
count			long	94
octave			long	7

l_vga6			long	@unpack_area	>> 2 << 4 |	0x0000		>> 2 << 18 |	6
l_vga7			long	@unpack_area	>> 2 << 4 |	0x0004		>> 2 << 18 |	7
l_synth			long	@unpack_area	>> 2 << 4 |	@channels	>> 2 << 18 |	5

black		long $02

vidpal			byte	$00, $01, $00, $01
palette			byte	$02, $07, $07, $07

threshold		long	40

			' Frequency base table (top octave).

freqbase		word	$1322, $1446, $157a, $16c1
			word	$181c, $198b, $1b0f, $1cab
			word	$1e60, $202e, $2218, $241f

message			byte	15, " This demo has"
			byte	"to be stored on"
			byte	"EEPROM. It uses"
			byte	"the entire RAM "
			byte	"for effects and"
			byte	"loads code over"
			byte	$49, $b2, "C as needed. "

			align	4
			pool

savereader		res	1
t2			res	1
txtaddr			res	1
x			res	1
y			res	1
bits			res	1

			vfit	496

