			include "propeller.i"
			include "demo.i"

			org	0

RATE			=	16

			' The smoke effect takes place in a 48 x 60 field, of which
			' the border values are zero all the time.

			' frontbuf and backbuf each point to 14848 ($3a00) bytes.
			' These are each divided into three sub-areas: 
			' At $0000: 48 x 60 bytes of unsigned 8-bit density ($b40)
			' At $0c00: 48 x 60 words of signed 16-bit x-velocity ($1680)
			' At $2300: 48 x 60 words of signed 16-bit y-velocity ($1680)

BUF_U			=	$0c00
BUF_V			=	$2300

HEIGHT			=	48

			mov	0, PAR		wz nr
		if_nz	jmp	#slave

			wrlong	frontback, =@backbuf
loop
			' Add static forces to the current (front) velocity field.

			mov	count, #46
			mov	addr, front_v
			add	addr, =(48 * (HEIGHT - 1) + 1) * 2
:forcelp
			wrword	minus328, addr
			add	addr, #2
			djnz	count, #:forcelp

			' Add buoyancy forces.

			mov	src_addr, front_d
			add	src_addr, =1
			mov	dest_addr, front_v
			add	dest_addr, =1 * 2
			mov	ycount, #HEIGHT
:ybuoy
			mov	xcount, #46
:xbuoy
			rdbyte	t1, src_addr
			add	src_addr, #1
			shr	t1, #1
			rdword	t2, dest_addr
			add	t2, #7
			sub	t2, t1
			wrword	t2, dest_addr
			add	dest_addr, #2
			djnz	xcount, #:xbuoy

			add	src_addr, #2
			add	dest_addr, #4
			djnz	ycount, #:ybuoy

			' Adjust the velocity field (front) using back-u/v as temporary storage.

			call	#project

			' Fill back-u/v with values from front-u/v advected by front-u/v.

			' This job can be parallellized: we delegate the y (v) coordinates to a slave cog.

			wrbyte	one, =@slave_enable

			mov	src_addr, front_u
			mov	dest_addr, back_u
			call	#advect16

:wait			rdbyte	t1, =@slave_enable
			tjnz	t1, #:wait

			' Swap the internal buffer pointers.

			rol	frontback, #16
			rol	back_d, #16
			rol	back_u, #16
			rol	back_v, #16
			rol	front_d, #16
			rol	front_u, #16
			rol	front_v, #16

			' Adjust the new velocity field (front) using back-u/v as temporary storage.

			call	#project

			' Let the new velocity field (front) affect the density field.
			' Fill front-d with values from back-d advected by front-u/v.

			mov	dest_addr, front_d
			mov	src_addr, back_d
			call	#advect

			mov	addr, front_d
			add	addr, =48 * (HEIGHT - 1) + 1
			mov	count, #46
:sourcelp
			mov	t2, lfsr
			shr	lfsr, #1	wc
		if_c	xor	lfsr, taps
			and	t2, #$7f
			mov	t1, #23
			sub	t1, count
			shl	t1, #3
			subabs	t2, t1
			mov	t1, source
			shr	t1, #1
			add	t2, t1
			sub	t2, =$1a8
			mins	t2, #0
			shl	t2, #1
			rdbyte	t1, addr
			add	t1, t2
			max	t1, #$c0
			wrbyte	t1, addr
			add	addr, #1
			djnz	count, #:sourcelp

			' Let the source decay.
			sub	source, #1

			' Swap the visible buffer.
			wrlong	frontback, =@backbuf

			cmps	source, =$d000 >> 6	wz wc
		if_b	mov	source, #0

			wrbyte	zero, =@internalvbl
:wvbl			rdbyte	t1, =@internalvbl
			tjz	t1, #:wvbl

			jmp	#loop

slave
			rdbyte	t1, =@slave_enable
			tjz	t1, #slave

			mov	src_addr, front_v
			mov	dest_addr, back_v
			call	#advect16

			rol	back_u, #16
			rol	back_v, #16
			rol	front_u, #16
			rol	front_v, #16

			wrbyte	zero, =@slave_enable
			jmp	#slave

project
			' Solve front-u/v for conservation of mass, which will introduce turbulence.

			' Initialize p (back_u).

			mov	addr, back_u
			mov	count, =HEIGHT * 48 / 2
:clrp
			wrlong	zero, addr
			add	addr, #4
			djnz	count, #:clrp

			' Initialize div (back_v).

			mov	addr, back_v

			mov	src_addr, front_v
			sub	src_addr, =(48 - 1) * 2		' point above the first pixel

			mov	ycount, #HEIGHT
:yinit
			mov	xcount, #46
:xinit
			rdword	t1, src_addr			' v above
			add	src_addr, =(48 + 48) * 2
			add	addr, #2
			rdword	t2, src_addr			' v below
			sub	t1, t2
			sub	src_addr, =BUF_V - BUF_U + (48 + 1) * 2
			rdword	t2, src_addr			' u left
			add	t1, t2
			add	src_addr, =2 * 2
			rdword	t2, src_addr			' u right
			sub	t1, t2
			shl	t1, #16
			sar	t1, #17
			wrword	t1, addr
			add	src_addr, =BUF_V - BUF_U - 48 * 2
			djnz	xcount, #:xinit

			add	addr, #4
			add	src_addr, #4
			djnz	ycount, #:yinit

			' Find the gradient using successive over-relaxation.

			mov	count, #20
:iterate
			mov	addr, back_u
			mov	ycount, #HEIGHT
:ysolve
			mov	t1, #0
			mov	xcount, #46
:xsolve
			'rdword	t1, addr			' p left
			add	addr, =BUF_V - BUF_U + 2
			rdword	t2, addr			' div (center)
			add	t1, t2
			sub	addr, =BUF_V - BUF_U + 48 * 2
			rdword	t2, addr			' p above
			add	t1, t2
			add	addr, =(48 + 1) * 2
			rdword	t2, addr			' p right
			add	t1, t2
			add	addr, =(48 - 1) * 2
			rdword	t2, addr			' p below
			add	t1, t2
			shl	t1, #16
			sar	t1, #18
			sub	addr, =48 * 2
			wrword	t1, addr			' write to p
			djnz	xcount, #:xsolve

			add	addr, #4
			djnz	ycount, #:ysolve

			djnz	count, #:iterate

			' Subtract the gradient from the velocity field.

			mov	dest_addr, front_v
			mov	addr, back_u
			mov	ycount, #HEIGHT
:yapply
			mov	xcount, #46
:xapply
			rdword	t1, addr			' p left
			sub	dest_addr, =BUF_V - BUF_U - 2	' goto next u
			add	addr, #2 * 2
			rdword	t2, addr			' p right
			sub	t1, t2
			shl	t1, #16
			rdword	t2, dest_addr
			sar	t1, #17
			add	t2, t1
			wrword	t2, dest_addr

			add	dest_addr, =BUF_V - BUF_U	' goto same v
			sub	addr, =(48 + 1) * 2
			rdword	t1, addr			' p above
			add	addr, =(48 + 48) * 2
			rdword	t2, addr			' p below
			sub	t1, t2
			shl	t1, #16
			rdword	t2, dest_addr
			sar	t1, #17
			add	t2, t1
			wrword	t2, dest_addr

			sub	addr, =48 * 2
			djnz	xcount, #:xapply

			add	dest_addr, #4
			add	addr, #4
			djnz	ycount, #:yapply

project_ret		ret

advect
			mov	fieldaddr, front_u
			add	fieldaddr, =1 * 2
			add	dest_addr, =1
			add	src_addr, =1
			mov	ycount, #HEIGHT
:yloop
			mov	xcount, #46
:xloop
			' Find the (x, y) coordinates (in fixed point 6.8) to fetch density from.

			rdword	t1, fieldaddr
			shl	t1, #16
			add	fieldaddr, =BUF_V - BUF_U
			rdword	t2, fieldaddr
			sar	t1, #RATE
			neg	x, t1
			shl	t2, #16
			sar	t2, #RATE
			neg	y, t2

			sub	fieldaddr, =BUF_V - BUF_U - 2

			' Extract the weights (fractional parts of x and y).

			mov	weightx, x
			and	weightx, #$7f
			mov	weighty, y
			and	weighty, #$7f
			sar	x, #7
			sar	y, #7

			' Calculate a weighted sum.

			mov	addr, src_addr
			shl	y, #4
			add	addr, y
			add	addr, y
			add	addr, y
			add	addr, x

			rdbyte	lerp_a, addr		' x0, y0
			add	addr, #48
			mov	mulu_s, weighty
			rdbyte	mulu_x, addr		' x0, y1

			sub	mulu_x, lerp_a

			abs	mulu_x, mulu_x		wc
			muxc	mulu_s, #1 << 7

			shl	mulu_x, #16
			shr	mulu_s, #1		wc
		if_c	add	mulu_s, mulu_x		wc
			rcr	mulu_s, #1		wc
		if_c	add	mulu_s, mulu_x		wc
			rcr	mulu_s, #1		wc
		if_c	add	mulu_s, mulu_x		wc
			rcr	mulu_s, #1		wc
		if_c	add	mulu_s, mulu_x		wc
			rcr	mulu_s, #1		wc
		if_c	add	mulu_s, mulu_x		wc
			rcr	mulu_s, #1		wc
		if_c	add	mulu_s, mulu_x		wc
			rcr	mulu_s, #1		wc
		if_c	add	mulu_s, mulu_x		wc
			rcr	mulu_s, #1		wc	' sign bit
			shr	mulu_s, #9

			negc	subsum, mulu_s

			shl	lerp_a, #7
			add	subsum, lerp_a

			add	subsum, #64
			shr	subsum, #7

			sub	addr, #48 - 1
			rdbyte	lerp_a, addr		' x1, y0
			add	addr, #48
			mov	mulu_s, weighty
			rdbyte	mulu_x, addr		' x1, y1

			sub	mulu_x, lerp_a

			abs	mulu_x, mulu_x		wc
			muxc	mulu_s, #1 << 7

			shl	mulu_x, #16
			shr	mulu_s, #1		wc
		if_c	add	mulu_s, mulu_x		wc
			rcr	mulu_s, #1		wc
		if_c	add	mulu_s, mulu_x		wc
			rcr	mulu_s, #1		wc
		if_c	add	mulu_s, mulu_x		wc
			rcr	mulu_s, #1		wc
		if_c	add	mulu_s, mulu_x		wc
			rcr	mulu_s, #1		wc
		if_c	add	mulu_s, mulu_x		wc
			rcr	mulu_s, #1		wc
		if_c	add	mulu_s, mulu_x		wc
			rcr	mulu_s, #1		wc
		if_c	add	mulu_s, mulu_x		wc
			rcr	mulu_s, #1		wc	' sign bit
			shr	mulu_s, #9

			negc	mulu_x, mulu_s

			shl	lerp_a, #7
			add	mulu_x, lerp_a

			add	mulu_x, #64
			shr	mulu_x, #7

			mov	mulu_s, weightx

			sub	mulu_x, subsum

			abs	mulu_x, mulu_x		wc
			muxc	mulu_s, #1 << 7

			shl	mulu_x, #16
			shr	mulu_s, #1		wc
		if_c	add	mulu_s, mulu_x		wc
			rcr	mulu_s, #1		wc
		if_c	add	mulu_s, mulu_x		wc
			rcr	mulu_s, #1		wc
		if_c	add	mulu_s, mulu_x		wc
			rcr	mulu_s, #1		wc
		if_c	add	mulu_s, mulu_x		wc
			rcr	mulu_s, #1		wc
		if_c	add	mulu_s, mulu_x		wc
			rcr	mulu_s, #1		wc
		if_c	add	mulu_s, mulu_x		wc
			rcr	mulu_s, #1		wc
		if_c	add	mulu_s, mulu_x		wc
			rcr	mulu_s, #1		wc	' sign bit
			shr	mulu_s, #9

			negc	mulu_s, mulu_s

			shl	subsum, #7
			add	mulu_s, subsum

			add	mulu_s, #13
			shr	mulu_s, #7

			add	src_addr, #1

			wrbyte	mulu_s, dest_addr

			add	dest_addr, #1
			djnz	xcount, #:xloop

			add	dest_addr, #2
			add	src_addr, #2
			add	fieldaddr, #4
			djnz	ycount, #:yloop

advect_ret		ret

advect16
			mov	fieldaddr, front_u
			add	fieldaddr, =1 * 2
			add	dest_addr, =1 * 2
			add	src_addr, =1 * 2

			mov	ycount, #HEIGHT
:yloop
			mov	xcount, #46
:xloop
			' Find the (x, y) coordinates (in fixed point 6.8) to fetch density from.

			rdword	t1, fieldaddr
			shl	t1, #16
			add	fieldaddr, =BUF_V - BUF_U
			rdword	t2, fieldaddr
			sar	t1, #RATE
			neg	x, t1
			shl	t2, #16
			sar	t2, #RATE
			neg	y, t2

			' Extract the weights (fractional parts of x and y).

			mov	weightx, x
			and	weightx, #$7f
			mov	weighty, y
			and	weighty, #$7f
			sar	x, #7
			sar	y, #7

			' Calculate a weighted sum.

			mov	addr, src_addr
			shl	y, #5
			add	addr, y
			add	addr, y
			add	addr, y
			add	addr, x
			add	addr, x

			rdword	lerp_a, addr		' x0, y0
			shl	lerp_a, #16
			add	addr, =48 * 2
			rdword	mulu_x, addr		' x0, y1
			sar	lerp_a, #16
			shl	mulu_x, #16
			sar	mulu_x, #16
			mov	mulu_s, weighty

			sub	mulu_x, lerp_a

			abs	mulu_x, mulu_x		wc
			muxc	mulu_s, #1 << 7

			shl	mulu_x, #16
			shr	mulu_s, #1		wc
		if_c	add	mulu_s, mulu_x		wc
			rcr	mulu_s, #1		wc
		if_c	add	mulu_s, mulu_x		wc
			rcr	mulu_s, #1		wc
		if_c	add	mulu_s, mulu_x		wc
			rcr	mulu_s, #1		wc
		if_c	add	mulu_s, mulu_x		wc
			rcr	mulu_s, #1		wc
		if_c	add	mulu_s, mulu_x		wc
			rcr	mulu_s, #1		wc
		if_c	add	mulu_s, mulu_x		wc
			rcr	mulu_s, #1		wc
		if_c	add	mulu_s, mulu_x		wc
			rcr	mulu_s, #1		wc	' sign bit
			shr	mulu_s, #9

			negc	subsum, mulu_s

			shl	lerp_a, #7
			add	subsum, lerp_a

			add	subsum, #64
			sar	subsum, #7

			sub	addr, =(48 - 1) * 2
			rdword	lerp_a, addr		' x1, y0
			shl	lerp_a, #16
			add	addr, =48 * 2
			rdword	mulu_x, addr		' x1, y1
			sar	lerp_a, #16
			shl	mulu_x, #16
			sar	mulu_x, #16
			mov	mulu_s, weighty

			sub	mulu_x, lerp_a

			abs	mulu_x, mulu_x		wc
			muxc	mulu_s, #1 << 7

			shl	mulu_x, #16
			shr	mulu_s, #1		wc
		if_c	add	mulu_s, mulu_x		wc
			rcr	mulu_s, #1		wc
		if_c	add	mulu_s, mulu_x		wc
			rcr	mulu_s, #1		wc
		if_c	add	mulu_s, mulu_x		wc
			rcr	mulu_s, #1		wc
		if_c	add	mulu_s, mulu_x		wc
			rcr	mulu_s, #1		wc
		if_c	add	mulu_s, mulu_x		wc
			rcr	mulu_s, #1		wc
		if_c	add	mulu_s, mulu_x		wc
			rcr	mulu_s, #1		wc
		if_c	add	mulu_s, mulu_x		wc
			rcr	mulu_s, #1		wc	' sign bit
			shr	mulu_s, #9

			negc	mulu_x, mulu_s

			shl	lerp_a, #7
			add	mulu_x, lerp_a

			add	mulu_x, #64
			sar	mulu_x, #7

			mov	mulu_s, weightx

			sub	mulu_x, subsum

			abs	mulu_x, mulu_x		wc
			muxc	mulu_s, #1 << 7

			shl	mulu_x, #16
			shr	mulu_s, #1		wc
		if_c	add	mulu_s, mulu_x		wc
			rcr	mulu_s, #1		wc
		if_c	add	mulu_s, mulu_x		wc
			rcr	mulu_s, #1		wc
		if_c	add	mulu_s, mulu_x		wc
			rcr	mulu_s, #1		wc
		if_c	add	mulu_s, mulu_x		wc
			rcr	mulu_s, #1		wc
		if_c	add	mulu_s, mulu_x		wc
			rcr	mulu_s, #1		wc
		if_c	add	mulu_s, mulu_x		wc
			rcr	mulu_s, #1		wc
		if_c	add	mulu_s, mulu_x		wc
			rcr	mulu_s, #1		wc	' sign bit
			shr	mulu_s, #9

			negc	mulu_s, mulu_s

			shl	subsum, #7
			add	mulu_s, subsum

			add	mulu_s, #39
			sar	mulu_s, #7

			add	src_addr, #2
			sub	fieldaddr, =BUF_V - BUF_U - 2

			wrword	mulu_s, dest_addr

			add	dest_addr, #2
			djnz	xcount, #:xloop

			add	dest_addr, #4
			add	src_addr, #4
			add	fieldaddr, #4
			djnz	ycount, #:yloop

advect16_ret		ret

one			long	1

frontback		word	@graphics0,			@graphics1
back_d			word	@graphics0 + 48,		@graphics1 + 48
back_u			word	@graphics0 + BUF_U + 48 * 2,	@graphics1 + BUF_U + 48 * 2
back_v			word	@graphics0 + BUF_V + 48 * 2,	@graphics1 + BUF_V + 48 * 2
front_d			word	@graphics1 + 48,		@graphics0 + 48
front_u			word	@graphics1 + BUF_U + 48 * 2,	@graphics0 + BUF_U + 48 * 2
front_v			word	@graphics1 + BUF_V + 48 * 2,	@graphics0 + BUF_V + 48 * 2

source			long	65535 >> 6

minus328		long	65536-328

lfsr			long	1
taps			long	$c0000031

			pool

t1			res	1
t2			res	1
count			res	1
addr			res	1
ycount			res	1
xcount			res	1

x			res	1
y			res	1
weightx			res	1
weighty			res	1

subsum			res	1

dest_addr		res	1
src_addr		res	1
fieldaddr		res	1

mulu_s			res	1
mulu_x			res	1

lerp_a			res	1

			vfit	496
