
			include "propeller.i"
			include "demo.i"

			org	0

' Input: NCHANNEL longs, 0mmmmmmmm_00000_0aaaaaaaa_0vvvvvvvv
'			a = amount of FM (amplitude of modulator)
'			m = FM multiplier (6.2 fixed point)
'			v = volume (amplitude of carrier)
'        NCHANNEL words, carrier frequencies

			long	0[256]

			mov	CNT, CNT
			add	CNT, deltacnt

			mov	CTRA, =%0_00110_000_00000000_000000_000_001011
			mov	CTRB, =%0_00110_000_00000000_000000_000_001010
			mov	DIRA, =%00000000_00000000_00001100_00000000

soundloop		' 50 instructions + channel loop

			mov	accum, #0
			mov	nchan, #6

			mov	addr, PAR
			movs	p_ph2, #phases
			movd	p_ph1, #phases
			' up to three wait slots here

channel_loop		' 108 instructions per channel once synchronized to the hub.

			rdlong	mulu_s, addr
			movs	volume, mulu_s
			add	addr, #CH_FREQ
			rdword	freq, addr
			shr	mulu_s, #9
			movs	amount, mulu_s
			shr	mulu_s, #14

			shl	freq, #16
p_ph1			add	000, freq
p_ph2			mov	t1, 000

			add	p_ph2, #1
			add	p_ph1, =1 << 9

			' We have the oscillator phase in t1. First, perform FM by adding amount * sin(phase * multiplier / 4)

			mov	mulu_x, t1
			shr	mulu_x, #16
			call	#mulu
			shr	mulu_s, #3+2

			' |sin(sin_a[12..0])| -> sin_a[15..0], (sin(sin_a[12..0]) >= 0) -> z

			test	mulu_s, =$0800		wc
			test	mulu_s, =$1000		wz
			and	mulu_s, =$07ff
		if_c	xor	mulu_s, =$07ff
			shl	mulu_s, #1
			or	mulu_s, =$e000
			rdword	mulu_x, mulu_s

			mov	mulu_s, amount
			call	#mulu
			negnz	mulu_s, mulu_s
			shl	mulu_s, #10			' This constant controls the range of the FM amount
			add	mulu_s, t1

			' The modulated phase is now in mulu_s.

			' Second, evaluate the waveform, mapping from phase (mulu_s, 32-bit unsigned) to value (-1 ^ !z * mulu_x[15..0]).

			shr	mulu_s, #19

			' |sin(sin_a[12..0])| -> sin_a[15..0], (sin(sin_a[12..0]) >= 0) -> z

			test	mulu_s, =$0800		wc
			test	mulu_s, =$1000		wz
			and	mulu_s, =$07ff
		if_c	xor	mulu_s, =$07ff
			shl	mulu_s, #1
			or	mulu_s, =$e000
			' two wait slots here
			rdword	mulu_x, mulu_s

			' We have the waveform magnitude in mulu_x[15..0]. Scale by volume.

			mov	mulu_s, volume
			call	#mulu
			negnz	mulu_s, mulu_s

			' Now we've got the signed, scaled waveform signal in mulu_s, in the range [-$ffffff, $ffffff]

			' Add this channel to the mix.

			add	accum, mulu_s

			add	addr, #CH_SIZEOF-CH_FREQ
			djnz	nchan, #channel_loop

			' The complete mix is now in the range ]-$600_0000, $600_0000[, so we'll scale it up to ]-$6000_0000, $6000_0000[.

			shl	accum, #4

			' The dry signal is in accum (signed), and spans almost the entire 32-bit range.
			' From this, and the head of the reverb buffer, we create a left and a right signal.
			' The reverb data (16-bit signed) will be scaled by 1/2 and added to the dry signal.

			' Read out the head of the reverb buffer.

:p_rev1			mov	right, 0
			mov	left, right
			and	left, =$ffff0000
			sar	left, #1
			shl	right, #16
			sar	right, #1

			' Mix the reverb output with the dry signal.

			add	left, accum
			add	right, accum

			' Write to the tail end of the reverb buffer, swapping left/right.

			mov	t1, left
			shr	t1, #16
:p_rev2			mov	0, t1

			mov	t1, right
			and	t1, =$ffff0000
:p_rev3			add	250, t1			' Introduce a small difference in the left and right
							' reverb delays for comb-filter spatialization.

			' Advance the reverb pointers.

			add	:p_rev1, #1
			andn	:p_rev1, =1 << 8
			add	:p_rev2, =1 << 9
			andn	:p_rev2, =1 << 17
			add	:p_rev3, =1 << 9
			andn	:p_rev3, =1 << 17

			' Finally, perform stereo widening.

			' Convert to average and difference, increase the difference, convert back.
			' Adjust the 32-bit signed values to PWM range.

			mov	diff, right		' diff = right - left
			sub	diff, left
			sar	right, #2		' right = left = left / 4 + right / 4
			sar	left, #2
			add	right, left
			mov	left, right
			shl	diff, #1
			add	right, diff
			sub	left, diff

			add	left, =$80000000
			add	right, =$80000000

			waitcnt	CNT, deltacnt
			mov	FRQA, left
			mov	FRQB, right

			sub	visualize, #4
			or	visualize, =$7f80	' Wraps from $7f7c to $7ffc.
			wrlong	left, visualize

			jmp	#soundloop


			' mulu_x[15..0] * mulu_s[7..0] -> mulu_s[23..0] (clobbers c, preserves z)
			' 4*21 cycles
mulu
			shl	mulu_x, #16
			shr	mulu_s, #1		wc
		if_c	add	mulu_s, mulu_x		wc
			rcr	mulu_s, #1		wc
		if_c	add	mulu_s, mulu_x		wc
			rcr	mulu_s, #1		wc
		if_c	add	mulu_s, mulu_x		wc
			rcr	mulu_s, #1		wc
		if_c	add	mulu_s, mulu_x		wc
			rcr	mulu_s, #1		wc
		if_c	add	mulu_s, mulu_x		wc
			rcr	mulu_s, #1		wc
		if_c	add	mulu_s, mulu_x		wc
			rcr	mulu_s, #1		wc
		if_c	add	mulu_s, mulu_x		wc
			rcr	mulu_s, #1		wc
		if_c	add	mulu_s, mulu_x		wc
			rcr	mulu_s, #1		wc
			shr	mulu_s, #8
mulu_ret		ret

deltacnt		long	80000000/SAMPLEFREQ

visualize		long	$7ffc

amount			long	0
volume			long	0

			pool

t1			res	1
t2			res	1

left			res	1
right			res	1

nchan			res	1

freq			res	1

diff			res	1

phases			res	6

accum			res	1

mulu_s			res	1
mulu_x			res	1

sin_a			res	1

addr			res	1

			vfit	496
